/* arp.c
 * prints out arp packets
 */

#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include "colorprint.h"

#define DIE(x) if(len<x){WARN printf("Unexpected end of packet")NORMAL;return;}

void doARPpacket(unsigned char *data, int len);

void doARPpacket(unsigned char *data, int len) {
	unsigned long *addrlong;
	unsigned short *nextword = (unsigned short *)data;
	unsigned char *nextbyte = data;
	int hln, pln, i, p;
	struct hostent *host;
	struct in_addr addr;
	
	HEAD printf("ARP HEADER:") NORMAL;
	DIE(2);
	TEXT printf("Hardware address space:");
	switch(*nextword) {
	case 0x0100:
		OK printf("Ethernet") NORMAL;
		break;
	default:
		WARN printf("Unknown %04X", ntohs(*nextword)) NORMAL;
		break;
	}
	nextword++; nextbyte+=2;
	DIE(4);
	TEXT printf("Protocol address space:");
	switch(*nextword) {
	case 0x0008:
		OK printf("IP") NORMAL;
		p=1;
		break;
	default:
		WARN printf("Unknown %04X", ntohs(*nextword)) NORMAL;
		OK printf("Go look it up in RFC 1700 and mail mike@earth.li") NORMAL;
		p=0;
		break;
	}
	nextword++; nextbyte+=2;

	DIE(6);
	TEXT printf("Hardware address len:");
	TESTPRINTu(*nextbyte==6,hln=(*nextbyte),"should be 6 for ethernet") NORMAL;
	nextbyte++;
	TEXT printf("Protocol address len:");
	TESTPRINTu(*nextbyte==4,pln=(*nextbyte),"should be 4 for IPv4") NORMAL;
	nextbyte++; nextword++;

	DIE(8);
	TEXT printf("Operation:");
	switch(*nextword) {
	case 0x0100:
		OK printf("request") NORMAL;
		break;
	case 0x0200:
		OK printf("reply") NORMAL;
		break;
	default:
		WARN printf("%04X (unknown)",ntohs(*nextword)) NORMAL;
		break;
	}
	nextword++; nextbyte+=2;

	DIE(8+hln);
	TEXT printf("Source hardware address: ");
	for(i=0;i<hln;i++) {
		printf("%02X",*nextbyte++);
		if(i<hln-1) printf(":");
	}
	NORMAL;

	DIE(8+hln+pln);
	TEXT printf("Source protocol address: ");
	addrlong = (unsigned long *)nextbyte;
	for(i=0;i<pln;i++) {
		printf("%d",*nextbyte++);
		if(i<pln-1) printf(".");
	}
	if(p && pln==4) {
		addr.s_addr = *addrlong;
		host = gethostbyaddr((char *)&addr,sizeof(struct in_addr),AF_INET);
	} else {
		host = NULL;
	}
	if(p && pln==4 && host) {
		OK printf("(%s)",host->h_name) NORMAL;
	} else {
		NORMAL;
	}

	DIE(8+2*hln+pln);
	TEXT printf("Target hardware address: ");
	for(i=0;i<hln;i++) {
		printf("%02X",*nextbyte++);
		if(i<hln-1) printf(":");
	}
	NORMAL;

	DIE(8+2*hln+2*pln);
	TEXT printf("Target protocol address: ");
	addrlong = (unsigned long *)nextbyte;
	for(i=0;i<pln;i++) {
		printf("%d",*nextbyte++);
		if(i<pln-1) printf(".");
	}
	if(p && pln==4) {
		addr.s_addr = *addrlong;
		host = gethostbyaddr((char *)&addr,sizeof(struct in_addr),AF_INET);
	} else {
		host = NULL;
	}
	if(p && pln==4 && host) {
		OK printf("(%s)",host->h_name) NORMAL;
	} else {
		NORMAL;
	}

	len-=8+2*hln+2*pln;
	if(len) {
		TEXT printf("Trailing data: ");
		hexlprint(nextbyte,len);
	}
}
