/* ip.c
	Takes an IP header and prints it out
	Note to self: may need to strip 44 byte ethernet header (?!)
*/

#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include "csum.h"
#include "icmp.h"
#include "udp.h"
#include "tcp.h"
#include "colorprint.h"

#define END(x)  if(packlen<x){WARN printf("Premeture end of packet") NORMAL;\
                              return protocol;}

int printIPheader (unsigned char *packet, int packlen, int *hlen, int *dlen);
void doIPpacket(unsigned char *data, int len);

void doIPpacket(unsigned char *data, int len) {
	int proto, hlen=0, dlen=0;

	switch(proto = printIPheader (data, len, &hlen, &dlen)) {
	case -1:
		WARN printf("Fragmented or incomplete packet.  Bailing out.") NORMAL;
		break;
	case 0:
		WARN printf("Protocol type 0 shouldn't happen.  Bailing out.")
			NORMAL;
		break;
	case 1:    /* ICMP */
		if(dlen > len) dlen = len;
		if(hlen < dlen) {
			printICMPpacket(data+hlen, dlen-hlen);
		} else {
			WARN printf("Packet data lost") NORMAL;
		}
		break;
	case 6:    /* TCP  */
		if(dlen > len) dlen = len;
		if(hlen < dlen) {
			printTCPpacket(data+hlen, dlen-hlen, data);
		} else {
			WARN printf("Packet data lost") NORMAL;
		}
		break;
	case 17:   /* UDP  */
		if(dlen > len) dlen = len;
		if(hlen < dlen) {
			printUDPpacket(data+hlen, dlen-hlen, data);
		} else {
			WARN printf("Packet data lost") NORMAL;
		}
		break;
	default:
		if(dlen > len) dlen = len;
		WARN printf("Sorry.  Don't know what protocol #%d is yet",proto)
			NORMAL;
		break;
	}
}


int printIPheader (unsigned char *packet, int packlen, int *hlen, int *dlen) {
	int header_len, /* header length in bytes */
		dgram_len,   /* datagram length in bytes */
		octets_left; /* octets of header left (for options) */

	int protocol=-1;   /* protocol (for return) */

	unsigned short check;
	unsigned char *nextbyte = (unsigned char *)packet;
	unsigned short *nextword = (unsigned short *)packet;
	unsigned long *nextlong = (unsigned long *)packet;

	struct protoent *proto;
	struct hostent *host;
	struct in_addr addr;
	
	if(packlen < 4) {
		WARN printf("Not enough of packet to be worth decoding") NORMAL;
		return -1;
	}
	HEAD printf("IP HEADER") NORMAL;
	TEXT printf("Version:");
	TESTPRINTd(((*nextbyte) >> 4) == 4, (*nextbyte)>>4, "should be 4") NORMAL;
	header_len = ((*nextbyte) & 0x0F) << 2;
	TEXT printf("Header length:");
	TESTPRINTd(header_len > 19, header_len, "should be at least 20") NORMAL;
	
	nextbyte++;
	TEXT printf("Service type:");
	switch((*nextbyte) >> 5) {  /* Top 3 bits are the precedence */
	case 7:
		WARN printf("Network control");
		break;
	case 6:
		WARN printf("Internetwork control");
		break;
	case 5:
		WARN printf("CRITIC/ECP");
		break;
	case 4:
		WARN printf("Flash Override");
		break;
	case 3:
		WARN printf("Flash");
		break;
	case 2:
		WARN printf("Immediate");
		break;
	case 1:
		OK printf("Priority");
		break;
	case 0:
		TEXT printf("Routine");
		break;
	default:
		WARN printf("!!!!");
		break;
	}
	TESTPRINTs(!((*nextbyte) & 0x10),"normal delay,","low delay,");
	TESTPRINTs(!((*nextbyte) & 0x08),"normal throughput,","high throughput,");
	NORMAL;
	TEXT printf("             ");
	TESTPRINTs(!((*nextbyte) & 0x04),"normal reliability,","high reliability,");
	TEXT printf("reserved =");
	TESTPRINTu(!((*nextbyte) & 0x03),(*nextbyte) & 0x03,"should be 0") NORMAL;
	
	nextbyte++;
	nextword++;
	TEXT printf("Datagram length: %u bytes",
					dgram_len = ntohs(*nextword)) NORMAL;
	
	END(6);
	nextbyte+=2;
	nextword++;
	nextlong++;
	TEXT printf("ID: 0x%04X",ntohs(*nextword)) NORMAL;

	END(8);
	nextbyte+=2;
	nextword++;
	*nextword = ntohs(*nextword);   /* put it in host byte order */
	TEXT printf("Flags:");
	if((*nextword) & 0x8000) { /* bit 2 of flags, bit 7 of byte */
		WARN printf("reserved,");
	}
	TESTPRINTs(!((*nextbyte) & 0x4000),"may fragment,","don't fragment,");
	TESTPRINTs(!((*nextbyte) & 0x2000),"last fragment","more fragments") NORMAL;
	TEXT printf("Fragment offset:");
	TESTPRINTu(!((*nextword)&0x1FFF),((*nextword)&0x1FFF) << 3,"bytes") NORMAL;

	END(10);
	
	if((*nextword)&0x1FFF) protocol=-1; else protocol=0;
	*nextword = htons(*nextword);  /* and put it back how we found it... */

	nextbyte+=2;
	nextword++;
	nextlong++;
	TEXT printf("TTL: %u",*nextbyte) NORMAL;

	nextbyte++;
	TEXT printf("Protocol:");
	proto = getprotobynumber(*nextbyte);
	if(proto == NULL) {
		WARN printf("%u (unknown).  Go look it up and add it to /etc/protocols",
						*nextbyte) NORMAL;
		OK printf("Oh, and if you find it, mail me (mike@earth.li) and let me know") NORMAL;
	} else {
		OK printf(proto->p_name) NORMAL;
	}
	if(protocol==0) protocol = *nextbyte;
	endprotoent();
	
	END(12);
	nextbyte++;
	nextword++;
	check = *nextword;
	*nextword=0;  /* zero checksum to compute what it should be */
	TEXT printf("IP Checksum:");
	if(header_len > packlen) {
		WARN printf("%04X (can't check - incomplete packet)",check) NORMAL;
	} else {
		PRINTCSUM(check,csum((unsigned short *)packet,header_len)) NORMAL;
	}
	*nextword = check; /* restore it so that udp,tcp etc can use it */

	END(16);
	nextbyte+=2;
	nextword++;
	nextlong++;
	addr.s_addr = *nextlong;
	host = gethostbyaddr((char *)&addr,sizeof(struct in_addr),AF_INET);
	TEXT printf("Source: %u.%u.%u.%u",*nextbyte,*(nextbyte+1),
			 *(nextbyte+2), *(nextbyte+3));
	if(host == NULL) {
		WARN printf("not found:");
		switch(h_errno) {
		case HOST_NOT_FOUND:
			WARN printf("Unknown host") NORMAL;
			break;
		case NO_ADDRESS:
			WARN printf("no address associated") NORMAL;
			break;
		case NO_RECOVERY:
			WARN printf("nameserver error") NORMAL;
			break;
		case TRY_AGAIN:
			WARN printf("temporoary error.  I should try again, but cba") NORMAL;
			break;
		default:
			WARN printf("errno: %d\n",h_errno) NORMAL;
			break;
		}
	} else {
		OK printf("(%s)",host->h_name) NORMAL;
	}

	END(20);
	nextbyte+=4;
	nextword+=2;
	nextlong++;
	addr.s_addr = *nextlong;
	host = gethostbyaddr((char *)&addr,sizeof(struct in_addr),AF_INET);
	TEXT printf("Destination: %u.%u.%u.%u",*nextbyte,*(nextbyte+1),
			 *(nextbyte+2), *(nextbyte+3));
	if(host == NULL) {
		WARN printf("not found:");
		switch(h_errno) {
		case HOST_NOT_FOUND:
			WARN printf("Unknown host") NORMAL;
			break;
		case NO_ADDRESS:
			WARN printf("no address associated") NORMAL;
			break;
		case NO_RECOVERY:
			WARN printf("nameserver error") NORMAL;
			break;
		case TRY_AGAIN:
			WARN printf("temporoary error.  I should try again, but cba") NORMAL;
			break;
		default:
			WARN printf("errno: %d",h_errno) NORMAL;
			break;
		}
	} else {
		OK printf("(%s)",host->h_name) NORMAL;
	}
	endhostent();

	*hlen = header_len;
	*dlen = dgram_len;

	nextbyte+=4;
	octets_left=header_len - 20;
	if((!octets_left) || packlen == header_len) return protocol;
	HEAD printf("IP OPTIONS:") NORMAL;
	while((octets_left > 0) && (nextbyte < packlen+packet)) {
		if(*nextbyte & 0x80)         /* first bit */
			TEXT printf("(copied to all frags)");
		switch(*nextbyte & 0x60) {   /* next 2 bits */
		case 0:
			OK printf("control:");
			break;
		case 1:
			WARN printf("(reserved):");
			break;
		case 2:
			OK printf("debug:");
			break;
		case 3:
			WARN printf("(reserved):");
			break;
		default:
			WARN printf("!!!!:");
			break;
		}
		switch(*nextbyte & 0x1F) {  /* rest of octet */

			/* TODO : many of these contain interesting info that could be
				dumped.  Read RFC791
			*/

		case 0:
			TEXT printf("End of options (padding)") NORMAL;
			octets_left--;
			nextbyte++;
			break;
		case 1:
			TEXT printf("Nop") NORMAL;
			break;
		case 2:
			WARN printf("Security (unused - see RFC791):");
			nextbyte++;
			TEXT printf("length");
			TESTPRINTu(*nextbyte == 11,*nextbyte,"should be 11") NORMAL;
			octets_left-=*nextbyte;
			octets_left-=2;
			nextbyte += (*nextbyte);
			nextbyte++;
			break;
		case 3:
			OK printf("Loose source routing") NORMAL;
			nextbyte++;
			octets_left-=(*nextbyte)+2;
			nextbyte += (*nextbyte)+1;
			break;
		case 4:
			TEXT printf("Time stamp (see RFC791)") NORMAL;
			nextbyte++;
			octets_left-=(*nextbyte)+2;
			nextbyte += (*nextbyte)+1;
			break;
		case 7:
			OK printf("Record route") NORMAL;
			nextbyte++;
			octets_left-=(*nextbyte)+2;
			nextbyte += (*nextbyte)+1;
			break;
		case 8:
			TEXT printf("Stream ID (see RFC791):");
			nextbyte++;
			TEXT printf("length");
			TESTPRINTu(*nextbyte==4,*nextbyte,"should be 4 - not dumping");
			TEXT printf(": ");
			if(*nextbyte==4) {
				int i;
				for(i=1;i<5;i++)
					TEXT printf("%02X",*(nextbyte+i));
				NORMAL;
			}
			octets_left-=(*nextbyte)+2;
			nextbyte += (*nextbyte)+1;
			break;
		case 9:
			OK printf("Strict source routing") NORMAL;
			nextbyte++;
			octets_left -= (*nextbyte)+2;
			nextbyte += (*nextbyte)+1;
			break;
		default:
			WARN printf("Unknown option #%u (length",*(nextbyte++));
			WARN printf("%u)",*nextbyte) NORMAL;
			octets_left -= (*nextbyte)+2;
			nextbyte += (*nextbyte)+1;
			break;
		}
	}
	return protocol;
}
